from markdown import Markdown

from mkapi.core.node import Node, get_node

converter = Markdown()


def get_html(node):
    if not isinstance(node, Node):
        node = get_node(node)
    markdown = node.get_markdown()
    html = converter.convert(markdown)
    node.set_html(html)
    return node.render()


def display(name):
    from IPython.display import HTML

    return HTML(get_html(name))


def filter(name):
    """
    Examples:
        >>> filter("a.b.c")
        ('a.b.c', [])
        >>> filter("a.b.c|upper|strict")
        ('a.b.c', ['upper', 'strict'])
    """
    index = name.find("|")
    if index == -1:
        return name, []
    name, filters = name[:index], name[index + 1 :]
    return name, filters.split("|")



"""This module provides entity classes to represent docstring structure."""
from dataclasses import dataclass, field
from typing import Iterator, List, Optional

import mkapi.core.preprocess
from mkapi.core import linker
from mkapi.core.preprocess import strip_ptags
from mkapi.core.regex import LINK_PATTERN
from mkapi.core.signature import Signature


@dataclass
class Base:
    """Base class.

    Args:
        name: Object name.
        markdown: Markdown source.

    Attributes:
        name: Object name.
        markdown: Markdown source.
        html: HTML after conversion.
    """

    name: str = ""
    markdown: str = ""
    html: str = field(default="", init=False)

    def set_html(self, html: str):
        """Sets `html` attribute.

        Args:
            html: HTML string.
        """
        self.html = html


@dataclass
class Type(Base):
    """Type class represents type for other classes."""

    def __post_init__(self):
        if not self.markdown:
            self.markdown = self.name

    def __bool__(self):
        return self.name != ""

    def __iter__(self) -> Iterator[Base]:
        if LINK_PATTERN.search(self.markdown):
            yield self

    def set_html(self, html: str):
        """Sets `html` attribute cleaning `p` tags."""
        self.html = strip_ptags(html)


@dataclass
class Item(Base):
    """Item class represents an item in Parameters, Attributes, and Raises sections."""

    type: Type = field(default_factory=Type)

    def __iter__(self) -> Iterator[Base]:
        yield from self.type
        yield self

    def set_html(self, html: str):
        """Sets `html` attribute cleaning `p` tags."""
        self.html = strip_ptags(html)


@dataclass
class Section(Base):
    """Section class represents a section in docstring.

    Args:
        items: List for Arguments, Attributes, or Raises sections.

    Attributes:
        items: List for Arguments, Attributes, or Raises sections.

    Examples:
        `Section` is iterable:
        >>> section = Section('Returns', markdown='An integer.')
        >>> for x in section:
        ...     assert x is section

        >>> items = [Item('x'), Item('y'), Item('z')]
        >>> section = Section('Parameters', items=items)
        >>> [item.name for item in section]
        ['x', 'y', 'z']

        Indexing:
        >>> isinstance(section['y'], Item)
        True
        >>> section['z'].name
        'z'

        Contains:
        >>> 'x' in section
        True
    """

    items: List[Item] = field(default_factory=list)
    type: Type = field(default_factory=Type)

    def __post_init__(self):
        if self.markdown:
            self.markdown = mkapi.core.preprocess.convert(self.markdown)

    def __iter__(self) -> Iterator[Base]:
        yield from self.type
        if self.markdown:
            yield self
        for item in self.items:
            yield from item

    def __getitem__(self, name) -> Optional[Item]:
        for item in self.items:
            if item.name == name:
                return item
        return None

    def __contains__(self, name) -> bool:
        return self[name] is not None


@dataclass
class Docstring:
    """Docstring class represents a docstring of an object.

    Args:
        sections: List of Section instance.
        type: Type for Returns and Yields sections.

    Attributes:
        sections: List of Section instance.
        type: Type for Returns and Yields sections.

    Examples:
        Empty docstring:
        >>> docstring = Docstring()
        >>> assert not docstring

        Docstring with 3 sections:
        >>> default = Section("", markdown="Default")
        >>> parameters = Section("Parameters", items=[Item("a"), Item("b")])
        >>> returns = Section("Returns", markdown="Results")
        >>> docstring = Docstring([default, parameters, returns])

        `Docstring` is iterable:
        >>> [base.name for base in docstring]
        ['', 'a', 'b', 'Returns']

        Indexing:
        >>> docstring["Parameters"].items[0].name
        'a'
    """

    sections: List[Section] = field(default_factory=list)
    type: Type = field(default_factory=Type)

    def __bool__(self):
        return len(self.sections) > 0

    def __iter__(self) -> Iterator[Base]:
        yield from self.type
        for section in self.sections:
            yield from section

    def __getitem__(self, name: str) -> Optional[Section]:
        for section in self.sections:
            if section.name == name:
                return section
        return None

    def __setitem__(self, name: str, section: Section):
        for k, section_ in enumerate(self.sections):
            if section_.name == name:
                self.sections[k] = section
                return
            if name == "Parameters" and section_.name != "":
                self.sections.insert(k, section)
                return
            if section_.name in ["Raises", "Yields", "Returns"]:
                self.sections.insert(k, section)
                return
        self.sections.append(section)


@dataclass
class Object(Base):
    """Object class represents an object."""

    prefix: str = ""
    id: str = field(init=False)
    kind: str = ""
    type: Type = field(default_factory=Type)
    signature: Signature = field(default_factory=Signature)

    def __post_init__(self):
        self.id = self.name
        if self.prefix:
            self.id = ".".join([self.prefix, self.name])
        if not self.markdown:
            name = linker.link(self.name, self.id)
            if self.prefix:
                prefix = linker.link(self.prefix, self.prefix)
                self.markdown = ".".join([prefix, name])
            else:
                self.markdown = name

    def __iter__(self) -> Iterator[Base]:
        yield from self.type
        yield self



import atexit
import logging
import os
import shutil
import sys
from typing import Dict, List, Tuple

from mkapi import utils
from mkapi.core.module import Module, get_module

logger = logging.getLogger("mkdocs")


def create_nav(config):
    nav = config["nav"]
    docs_dir = config["docs_dir"]
    config_dir = os.path.dirname(config["config_file_path"])
    abs_api_paths = []
    for page in nav:
        if isinstance(page, dict):
            for key, value in page.items():
                if isinstance(value, str) and value.startswith("mkapi/"):
                    page[key], abs_api_paths_ = collect(value, docs_dir, config_dir)
                    abs_api_paths.extend(abs_api_paths_)
    return config, abs_api_paths


def collect(path: str, docs_dir: str, config_dir) -> Tuple[list, list]:
    _, api_path, *paths, package_path = path.split("/")
    abs_api_path = os.path.join(docs_dir, api_path)
    if os.path.exists(abs_api_path):
        logger.error(f"[MkApi] {abs_api_path} exists: Delete manually for safety.")
        sys.exit(1)
    os.mkdir(abs_api_path)
    atexit.register(lambda path=abs_api_path: rmtree(path))

    root = os.path.join(config_dir, *paths)
    if root not in sys.path:
        sys.path.insert(0, root)

    package_path, filters = utils.filter(package_path)

    module = get_module(package_path)
    nav = []
    abs_api_paths = []
    modules: Dict[str, str] = {}
    package = None

    def add_page(module: Module) -> str:
        page_file = module.object.id + ".md"
        abs_path = os.path.join(abs_api_path, page_file)
        abs_api_paths.append(abs_path)
        create_page(abs_path, module, filters)
        return os.path.join(api_path, page_file)

    for m in module:
        if m.object.kind == "package":
            if package and modules:
                nav.append({package: modules})
            package = m.object.id
            modules = {}
            if m.docstring:
                modules[m.object.id] = add_page(m)
        else:
            modules[m.object.name] = add_page(m)
    if package and modules:
        nav.append({package: modules})

    return nav, abs_api_paths


def create_page(path: str, module: Module, filters: List[str]):
    with open(path, "w") as f:
        f.write(module.get_markdown(filters))


def rmtree(path: str):
    if not os.path.exists(path):
        return
    try:
        shutil.rmtree(path)
    except PermissionError:
        logger.warning(f"[MkApi] Couldn't delete directory: {path}")



import os
from dataclasses import dataclass, field
from typing import Dict, List

from jinja2 import Environment, FileSystemLoader, Template, select_autoescape

import mkapi
from mkapi.core import linker


@dataclass
class Renderer:
    templates: Dict[str, Template] = field(default_factory=dict, init=False)
    upper: bool = False

    def __post_init__(self):
        path = os.path.join(os.path.dirname(mkapi.__file__), "templates")
        loader = FileSystemLoader(path)
        env = Environment(loader=loader, autoescape=select_autoescape(["jinja2"]))
        for name in os.listdir(path):
            template = env.get_template(name)
            name = os.path.splitext(name)[0]
            self.templates[name] = template

    def render(self, node) -> str:
        object = self.render_object(node.object, heading=node.parent is None)
        docstring = self.render_docstring(node.docstring)
        members = []
        if node.members:
            members = [self.render(member) for member in node.members]
        return self.render_node(node, object, docstring, members)

    def render_node(self, node, object, docstring: str, members: List[str]) -> str:
        template = self.templates["node"]
        return template.render(
            node=node, object=object, docstring=docstring, members=members
        )

    def render_object(self, object, heading):
        context = linker.resolve_object(object.html)
        if context["level"] and heading:
            template = self.templates["object_heading"]
        else:
            template = self.templates["object_div"]
        html = template.render(context, object=object, upper=self.upper and heading)
        return html

    def render_docstring(self, docstring) -> str:
        if docstring is None:
            return ""
        template = self.templates["docstring"]
        for section in docstring.sections:
            if section.items:
                section.html = self.render_section(section)
        return template.render(docstring=docstring)

    def render_section(self, section) -> str:
        if section.name in ["Parameters", "Attributes", "Raises"]:
            return self.templates["args"].render(section=section)
        else:
            raise ValueError(f"Invalid section name: {section.name}")

    def render_module(self, module, filters) -> str:
        module_filter = ""
        if "upper" in filters:
            module_filter = "|upper"
            filters = filters.copy()
            del filters[filters.index("upper")]
        object_filter = "|" + "|".join(filters)
        template = self.templates["module"]
        return template.render(
            module=module, module_filter=module_filter, object_filter=object_filter
        )


renderer = Renderer()



import re

LINK_PATTERN = re.compile(r"\[(.+?)\]\((.+?)\)")

MKAPI_PATTERN = re.compile(r"^(#*) *?!\[mkapi\]\((.+?)\)$", re.MULTILINE)

NODE_PATTERN = re.compile(
    r"<!-- mkapi:begin:(\d+):(\w+) -->(.*?)<!-- mkapi:end -->", re.MULTILINE | re.DOTALL
)


def node_markdown(index: int, markdown: str, upper: bool = False) -> str:
    return f"<!-- mkapi:begin:{index}:{upper} -->\n\n{markdown}\n\n<!-- mkapi:end -->"



from dataclasses import dataclass, field
from typing import Any, List

from mkapi.core.base import Object
from mkapi.core.docstring import Docstring, get_docstring
from mkapi.core.object import get_sourcefile_and_lineno, split_prefix_and_name
from mkapi.core.signature import get_signature


@dataclass
class Tree:
    """Tree class. This class is the base of Node and Module.

    Attributes:
        obj: Object.
        prefix: Prefix.
        name: Name.
        id: ID for CSS.
        kind: Kind such as `function`, `class`, `module`, etc.
        docstring: Docstring instance.
        sourcefile: Source file path.
        lineno: int: Line number.
        members: Member objects. For example, methods of class.
        prefix_url: URL for prefix link.
        name_url: URL for name link.
    """

    obj: Any = field(repr=False)
    sourcefile: str = field(init=False)
    lineno: int = field(init=False)
    object: Object = field(init=False)
    docstring: Docstring = field(init=False)
    parent: Any = field(default=None, init=False)
    members: List[Any] = field(init=False)

    def __post_init__(self):
        obj = self.obj
        self.sourcefile, self.lineno = get_sourcefile_and_lineno(obj)
        prefix, name = split_prefix_and_name(obj)
        kind = self.get_kind()
        signature = get_signature(obj)
        self.object = Object(prefix=prefix, name=name, kind=kind, signature=signature)
        self.docstring = get_docstring(obj)
        self.members = self.get_members()
        for member in self.members:
            member.parent = self

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.members[index]
        else:
            for member in self.members:
                if member.object.name == index:
                    return member

    def __getattr__(self, name):
        return self[name]

    def __len__(self):
        return len(self.members)

    def __contains__(self, name):
        for member in self.members:
            if member.object.name == name:
                return True
        return False

    def get_kind(self) -> str:
        """Returns kind of this object."""
        raise NotImplementedError

    def get_members(self) -> List["Tree"]:
        """Returns a list of members."""
        raise NotImplementedError

    def get_markdown(self) -> str:
        """Returns a Markdown source for docstring of this object."""
        raise NotImplementedError



__version__ = "0.7.1"

from mkapi.core.module import get_module
from mkapi.core.node import get_node
from mkapi.utils import display, get_html

__all__ = ["get_node", "get_module", "get_html", "display"]



import importlib
import inspect
from typing import Any, Tuple


def get_object(name: str) -> Any:
    """Reutrns an object specified by `name`.

    Args:
        name: Object name.

    Examples:
        >>> import inspect
        >>> obj = get_object('mkapi.core')
        >>> inspect.ismodule(obj)
        True
        >>> obj = get_object('mkapi.core.base')
        >>> inspect.ismodule(obj)
        True
        >>> obj = get_object('mkapi.core.node.Node')
        >>> inspect.isclass(obj)
        True
        >>> obj = get_object('mkapi.core.node.Node.get_markdown')
        >>> inspect.isfunction(obj)
        True
    """
    names = name.split(".")
    for k in range(len(names), 0, -1):
        module_name = ".".join(names[:k])
        try:
            obj = importlib.import_module(module_name)
        except ModuleNotFoundError:
            continue
        for attr in names[k:]:
            obj = getattr(obj, attr)
        return obj
    raise ValueError(f"Could not find object: {name}")


def split_prefix_and_name(obj) -> Tuple[str, str]:
    """Split an object full name into prefix and name.

    Examples:
        >>> import inspect
        >>> obj = get_object('mkapi.core')
        >>> split_prefix_and_name(obj)
        ('mkapi', 'core')
        >>> obj = get_object('mkapi.core.base')
        >>> split_prefix_and_name(obj)
        ('mkapi.core', 'base')
        >>> obj = get_object('mkapi.core.node.Node')
        >>> split_prefix_and_name(obj)
        ('mkapi.core.node', 'Node')
        >>> obj = get_object('mkapi.core.node.Node.get_markdown')
        >>> split_prefix_and_name(obj)
        ('mkapi.core.node.Node', 'get_markdown')
    """
    if isinstance(obj, property):
        obj = obj.fget
    if inspect.ismodule(obj):
        prefix, _, name = obj.__name__.rpartition(".")
    else:
        module = obj.__module__
        qualname = obj.__qualname__
        if "." not in qualname:
            prefix, name = module, qualname
        else:
            prefix, _, name = qualname.rpartition(".")
            prefix = ".".join([module, prefix])
        if prefix == "__main__":
            prefix = ""
    return prefix, name


def get_sourcefile_and_lineno(obj) -> Tuple[str, int]:
    if isinstance(obj, property):
        obj = obj.fget
    try:
        sourcefile = inspect.getsourcefile(obj) or ""
    except TypeError:
        sourcefile = ""
    try:
        lineno = inspect.getsourcelines(obj)[1]
    except (TypeError, OSError):
        lineno = -1
    return sourcefile, lineno



import inspect
import os
from dataclasses import dataclass, field
from typing import Iterator, List

from mkapi.core.node import get_kind
from mkapi.core.object import get_object, get_sourcefile_and_lineno
from mkapi.core.renderer import renderer
from mkapi.core.tree import Tree


@dataclass
class Module(Tree):
    """Module class represents an module.

    Attributes:
        kind: Kind. `package` or `module`.
    """

    members: List["Module"] = field(init=False)
    objects: List[str] = field(default_factory=list, init=False)

    def __post_init__(self):
        super().__post_init__()
        if self.object.kind == "module":
            objects = get_objects(self.obj)
            self.objects = [".".join([self.object.id, obj]) for obj in objects]

    def __iter__(self) -> Iterator["Module"]:
        if self.docstring:
            yield self
        elif self.object.kind == "package" and any(m.docstring for m in self.members):
            yield self
        for member in self.members:
            yield from member

    def get_kind(self) -> str:
        if not self.sourcefile or self.sourcefile.endswith("__init__.py"):
            return "package"
        else:
            return "module"

    def get_members(self) -> List["Module"]:  # type:ignore
        if self.object.kind == "module":
            return []
        else:
            return get_members(self.obj)

    def get_markdown(self, filters: List[str]) -> str:  # type:ignore
        """Returns a Markdown source for docstring of this object."""
        return renderer.render_module(self, filters)


def get_objects(obj) -> List[str]:
    obj_source_file = inspect.getsourcefile(obj)
    members = []
    for name, obj in inspect.getmembers(obj):
        if name.startswith("_"):
            continue
        if not get_kind(obj):
            continue
        sourcefile, lineno = get_sourcefile_and_lineno(obj)
        if sourcefile != obj_source_file:
            continue
        if not inspect.getdoc(obj):
            continue
        members.append((name, lineno))
    members = sorted(members, key=lambda x: x[1])
    return [x[0] for x in members]


def get_members(obj) -> List[Module]:
    sourcefile = inspect.getsourcefile(obj)
    if not sourcefile:
        return []
    root = os.path.dirname(sourcefile)
    paths = [path for path in os.listdir(root) if not path.startswith("_")]
    members = []
    for path in paths:
        root_ = os.path.join(root, path)
        name = ""
        if os.path.isdir(root_) and "__init__.py" in os.listdir(root_):
            name = path
        elif path.endswith(".py"):
            name = path[:-3]
        if name:
            name = ".".join([obj.__name__, name])
            module = get_module(name)
            members.append(module)
    return members


def get_module(name) -> Module:
    if isinstance(name, str):
        obj = get_object(name)
    else:
        obj = name

    return Module(obj)



import os
import re
from html.parser import HTMLParser
from typing import Any, Dict, List

from mkapi.core.regex import LINK_PATTERN


def link(name: str, href: str) -> str:
    return f"[{name}](!{href})"


def resolve_link(markdown: str, abs_src_path: str, abs_api_paths: List[str]) -> str:
    def replace(match):
        name, href = match.groups()
        if href.startswith("!!"):  # Just for MkApi documentation.
            href = href[2:]
            return f"[{name}]({href})"
        if href.startswith("!"):
            href = href[1:]
            from_mkapi = True
        else:
            from_mkapi = False

        href = resolve_href(href, abs_src_path, abs_api_paths)
        if href:
            return f"[{name}]({href})"
        elif from_mkapi:
            return name
        else:
            return match.group()

    return re.sub(LINK_PATTERN, replace, markdown)


def resolve_href(name: str, abs_src_path: str, abs_api_paths: List[str]) -> str:
    if not name:
        return ""
    abs_api_path = match_last(name, abs_api_paths)
    if not abs_api_path:
        return ""
    relpath = os.path.relpath(abs_api_path, os.path.dirname(abs_src_path))
    relpath = relpath.replace("\\", "/")
    return "#".join([relpath, name])


def match_last(name: str, abs_api_paths: List[str]) -> str:
    match = ""
    for abs_api_path in abs_api_paths:
        dirname, path = os.path.split(abs_api_path)
        if name.startswith(path[:-3]):
            match = abs_api_path
    return match


class ObjectParser(HTMLParser):
    def feed(self, html):
        self.context = {"href": [], "heading_id": ""}
        super().feed(html)
        href = self.context["href"]
        if len(href) == 2:
            prefix_url, name_url = href
        elif len(href) == 1:
            prefix_url, name_url = "", href[0]
        else:
            prefix_url, name_url = "", ""
        self.context["prefix_url"] = prefix_url
        self.context["name_url"] = name_url
        del self.context["href"]
        return self.context

    def handle_starttag(self, tag, attrs):
        context = self.context
        if tag == "p":
            context["level"] = 0
        elif re.match(r"h[1-6]", tag):
            context["level"] = int(tag[1:])
            for attr in attrs:
                if attr[0] == "id":
                    self.context["heading_id"] = attr[1]
        elif tag == "a":
            for attr in attrs:
                if attr[0] == "href":
                    href = attr[1]
                    if href.startswith("./"):
                        href = href[2:]
                    self.context["href"].append(href)


parser = ObjectParser()


def resolve_object(html: str) -> Dict[str, Any]:
    parser.reset()
    return parser.feed(html)



import inspect
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Any, Iterator, List

from mkapi.core.base import Base, Object
from mkapi.core.object import get_object
from mkapi.core.renderer import renderer
from mkapi.core.tree import Tree


@dataclass
class Node(Tree):
    """Node class represents an object.

    Attributes:
        obj: Object.
        signature: Signature instance.
        object: Object instance.
        docstring: Docstring instance.
        members: Member objects. For example, methods of class.
        html: HTML after rendering.
    """

    members: List["Node"] = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        members = self.members
        if self.object.kind in ["class", "dataclass"] and not self.docstring:
            for member in members:
                if member.object.name == "__init__" and member.docstring:
                    markdown = member.docstring.sections[0].markdown
                    if not markdown.startswith("Initialize self"):
                        self.docstring = member.docstring
        self.members = [m for m in members if m.object.name != "__init__"]
        if self.docstring and self.docstring.type:
            self.object.type = self.docstring.type

    def __iter__(self) -> Iterator[Base]:
        yield self.object
        yield from self.docstring
        for member in self.members:
            yield from member

    def get_kind(self) -> str:
        if inspect.ismodule(self.obj):
            if self.sourcefile.endswith("__init__.py"):
                return "package"
            else:
                return "module"
        return get_kind(self.obj)

    def get_members(self) -> List["Node"]:  # type:ignore
        return get_members(self.obj)

    def get_markdown(self, level: int = 0) -> str:
        """Returns a Markdown source for docstring of this object."""
        markdowns = []
        for base in self:
            markdown = base.markdown
            if isinstance(base, Object):
                if level and self.parent is None:
                    markdown = "#" * level + " " + markdown
            markdowns.append(markdown)
        return "\n\n<!-- mkapi:sep -->\n\n".join(markdowns)

    def set_html(self, html: str):
        """Sets HTML to `Base` instances recursively."""
        for base, html in zip(self, html.split("<!-- mkapi:sep -->")):
            base.set_html(html.strip())

    def render(self) -> str:
        self.html = renderer.render(self)
        return self.html


def get_kind(obj) -> str:
    if isinstance(obj, property):
        if obj.fset:
            return "readwrite_property"
        else:
            return "readonly_property"
    if hasattr(obj, "__dataclass_fields__") and hasattr(obj, "__qualname__"):
        return "dataclass"
    if inspect.isclass(obj):
        return "class"
    if inspect.isgeneratorfunction(obj):
        return "generator"
    if inspect.isfunction(obj):
        try:
            parameters = inspect.signature(obj).parameters
        except (ValueError, TypeError):
            return ""
        if parameters:
            arg = list(parameters)[0]
            if arg == "self":
                return "method"
        return "function"
    return ""


def is_member(name: str, obj: Any, sourcefile: str) -> bool:
    if isinstance(obj, property):
        return True
    if name.startswith("_") and name != "__init__":
        return False
    if not get_kind(obj):
        return False
    try:
        source = inspect.getsourcefile(obj)
    except TypeError:
        return False
    if sourcefile and source != sourcefile:
        return False
    return True


def get_members(obj) -> List[Node]:
    if inspect.ismodule(obj) or isinstance(obj, property):
        return []

    try:
        sourcefile = inspect.getsourcefile(obj) or ""
    except TypeError:
        sourcefile = ""
    members = []
    for name, obj in inspect.getmembers(obj):
        if is_member(name, obj, sourcefile):
            member = get_node(obj)
            if member.docstring:
                members.append(member)
    return sorted(members, key=lambda x: x.lineno)


@lru_cache(maxsize=1000)
def _get_node(obj) -> Node:
    return Node(obj)


def get_node(name, cache=True) -> Node:
    if isinstance(name, str):
        obj = get_object(name)
    else:
        obj = name

    if cache:
        return _get_node(obj)
    else:
        return Node(obj)



import re
from dataclasses import dataclass, field
from typing import Iterator, List

from mkapi.core.inherit import inherit
from mkapi.core.linker import resolve_link
from mkapi.core.node import Node, get_node
from mkapi.core.regex import MKAPI_PATTERN, NODE_PATTERN, node_markdown
from mkapi.core.renderer import renderer
from mkapi import utils


@dataclass
class Page:
    source: str
    abs_src_path: str
    abs_api_paths: List[str] = field(default_factory=list)
    markdown: str = field(init=False)
    nodes: List[Node] = field(default_factory=list, init=False)

    def __post_init__(self):
        markdown = "\n\n".join(self.split(self.source))
        self.markdown = resolve_link(markdown, self.abs_src_path, self.abs_api_paths)

    def split(self, source) -> Iterator[str]:
        cursor = 0
        for index, match in enumerate(MKAPI_PATTERN.finditer(source)):
            start, end = match.start(), match.end()
            if cursor < start:
                markdown = source[cursor:start].strip()
                if markdown:
                    yield markdown
            heading, name = match.groups()
            name, filters = utils.filter(name)
            node = get_node(name, cache='nocache' not in filters)
            if node.object.kind in ["class", "dataclass"]:
                if "inherit" in filters:
                    inherit(node)
                elif "strict" in filters:
                    inherit(node, strict=True)
            self.nodes.append(node)
            markdown = node.get_markdown(level=len(heading))
            yield node_markdown(index, markdown, 'upper' in filters)
            cursor = end
        if cursor < len(source):
            markdown = source[cursor:].strip()
            if markdown:
                yield markdown

    def content(self, html):
        def replace(match):
            node = self.nodes[int(match.group(1))]
            node.set_html(match.group(3))
            renderer.upper = match.group(2) == "True"
            return renderer.render(node)

        return re.sub(NODE_PATTERN, replace, html)
